# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
from collections import defaultdict

from lxml import etree


def iter_2d_dict(dct):
    for i, row in sorted(dct.items()):
        cols = []
        for j, col in sorted(row.items()):
            cols.append(col)
        yield cols


def table_to_list(table):
    return list(iter_2d_dict(table_to_2d_dict(table)))


def table_to_2d_dict(table: etree.ElementBase):
    result: dict[str, str] = defaultdict(lambda : defaultdict(str))
    for row_i, row in enumerate(table.xpath('./tr')):
        for col_i, col in enumerate(row.xpath('./td|./th')):
            colspan = int(col.get('colspan', 1))
            rowspan = int(col.get('rowspan', 1))
            col_data = str.strip(col.text_content())
            while row_i in result and col_i in result[row_i]:
                col_i += 1
            for i in range(row_i, row_i + rowspan):
                for j in range(col_i, col_i + colspan):
                    result[i][j] = col_data
    return result


# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import re
from types import ModuleType
from typing import Any
from typing import AsyncIterator
from typing import Callable

from headless.types import IClient

from .const import DEFAULT_GRADINGS
from .const import APPLE_MODELS
from . import v1
from . import v2
from . import v3


async def load_models(
    client: IClient[Any, Any],
    using: ModuleType
) -> AsyncIterator[dict[str, int | str]]:
    async for params in using.parse_iphones(client):
        yield params

    async for params in using.parse_ipads(client):
        # Create a variant for cellular, wifi
        for cap in {'C', 'W'}:
            params: dict[str, Any] = {
                **copy.deepcopy(params),
                'cellular': cap == 'C',
                'sku': params['sku'] + cap
            }
            yield params


def generate_name(params: dict[str, Any]) -> str:
    kind = str.lower(params.get('kind') or '')

    def generate_iphone_name(attrs: dict[str, Any]):
        name = '{model} {memory} {color}'.format(**attrs)
        return name


    def generate_ipad_name(attrs: dict[str, Any]):
        name = '{device} ({year})'.format(**attrs)
        if attrs.get('screen_size'):
            name = f'{name} {attrs["screen_size"]}"'
        name = f'{name} {"WiFi + Cellular" if attrs["cellular"] else "WiFi"}'
        name = '{name} {memory} {color}'.format(name=name, **params)
        return name

    name = None
    match kind:
        case "ipad":
            name = generate_ipad_name(params)
        case "iphone":
            name = generate_iphone_name(params)
        case _:
            raise NotImplementedError

    assert name is not None
    params.setdefault('grade_name', '')
    return str.strip('{name} {grade_name}'.format(name=name, **params))


async def generate(
    client: IClient[Any, Any],
    using: ModuleType,
    gradings: dict[str, Any] = DEFAULT_GRADINGS,
    generate_sku: Callable[[str, dict[str, Any]], str] | None = None,
    with_ungraded: bool = True
) -> AsyncIterator[dict[str, int | str]]:
    async for product in load_models(client, using=using):
        if gradings:
            for grade, sku in gradings.items():
                params = {
                    **copy.deepcopy(product),
                    'grading_level': 'ABC',
                    'grade': sku,
                    'grade_name': grade,
                    'sku': str(product['sku']) + str(sku)
                }
                yield {**params, 'name': generate_name(params)}

        params = copy.deepcopy(product)
        if with_ungraded:
            yield {**params, 'name': generate_name(params)}




model_pattern: re.Pattern[str] = re.compile(
    '(.*)(' + str.join('|', [
        x for x in reversed(APPLE_MODELS.keys())
    ]) + ')+'
)


def get_model_info(value: str) -> dict[str, Any]:
    m = model_pattern.match(value)
    if not m:
        raise ValueError("Unable to parse model")
    return APPLE_MODELS[m.group(2)]


# Copyright (C) 2022 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from datetime import datetime

import aorta
from canonical import EmailAddress

from ..devicesource import DeviceSource
from ..documentreference import DocumentReference
from ..imei import IMEI
from ..trackedmobiledevicereference import TrackedMobileDeviceReference


class DeviceReceived(aorta.Event):
    device: TrackedMobileDeviceReference
    timestamp: datetime
    document: DocumentReference
    handler: EmailAddress
    imei: list[IMEI]
    product_id: int | None = None
    sku: str | None = None
    source: DeviceSource


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Sequence
from typing import TypeVar

import pydantic


T = TypeVar('T', bound='SpreadsheetModel')


class SpreadsheetModel(pydantic.BaseModel):

    @classmethod
    def parse_row(
        cls: type[T],
        row: Sequence[Any]
    ) -> T:
        values: dict[str, Any] = {}
        for i, k in enumerate(cls.__fields__):
            try:
                values[k] = row[i]
            except IndexError:
                values[k] = None
        return cls.parse_obj(values)

    def as_row(self) -> dict[str, Any]:
        values = tuple(getattr(self, k) for k in self.__fields__.keys())
        return {'values': [
            {'userEnteredValue': {'stringValue': str(x or '')}}
            for x in values
        ]}


# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import re
from typing import Any

from .const import APPLE_MODELS
from .v2 import APPLE_COLORS


_color_pattern: re.Pattern[str] = re.compile(
    '(.*?)(' + str.join('|', [
        x for x in sorted(APPLE_COLORS.keys(), key=lambda x: -len(x))
    ]) + ')',
    re.IGNORECASE
)


_model_pattern: re.Pattern[str] = re.compile(
    '(.*?)(' + str.join('|', [
        x for x in sorted(APPLE_MODELS.keys(), key=lambda x: -len(x))
    ]) + ')',
    re.IGNORECASE
)


_storage_pattern: re.Pattern[str] = re.compile(
    r'(.*?)(([0-9]+(\s+)?)(GB|TB))',
    re.IGNORECASE
)


def parse_color(value: str) -> tuple[str, str] | tuple[None, None]:
    m = _color_pattern.match(value)
    if not m:
        return None, None
    color = m.group(2)
    return APPLE_COLORS[str.lower(color)], color


def parse_model(value: str) -> str | None:
    if 'iphone' in str.lower(value):
        m = _model_pattern.match(value)
        if not m:
            return None
        return m.group(2)
    else:
        return None


def parse_storage(value: str) -> str | None:
    m = _storage_pattern.match(value)
    if not m:
        return None
    storage = m.group(2)
    return str.upper(re.sub(r'\s+', '', storage))


def parse(value: str) -> dict[str, Any]:
    params: dict[str, Any] = {}
    params['color_code'], params['color'] = parse_color(value)
    params['model'] = model = parse_model(value)
    params['storage'] = parse_storage(value)
    if model:
        params.update(APPLE_MODELS[str.lower(model)])
    return params


# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from headless.types import IClient
from . import extract
from .const import APPLE_MODELS


APPLE_COLORS: dict[str, str] = {
    'Alpine Green'  : 'GR',
    'Black'         : 'B',
    'Black & Slate' : 'BS',
    'Blue'          : 'BL',
    'Coral'         : 'C',
    'Deep Purple'   : 'P',
    'Graphite'      : 'G',
    'Green'         : 'GR',
    'Gold'          : 'GO',
    'Jet Black'     : 'JB',
    'Midnight'      : 'M',
    'Midnight Green': 'MG',
    'Pink'          : 'PI',
    'Red'           : 'RE',
    'Rose Gold'     : 'GO',
    'PRODUCT(RED)'  : 'R',
    'Pacific Blue'  : 'BL',
    'Purple'        : 'P',
    'Sierra Blue'   : 'BL',
    'Silver'        : 'S',
    'Space Black'   : 'SBL',
    'Space Gray'    : 'SG',
    'Sky Blue'      : 'SK',
    'Starlight'     : 'ST',
    'White'         : 'W',
    'White & Silver': 'WS',
    'Yellow'        : 'Y',
}

GRADES: dict[str, str] = {
    'A+'    : '0',
    'A'     : '1',
    'B+'    : '2',
    'B'     : '3',
    'C+'    : '4',
    'C'     : '5',
    'D'     : '6'
}
GRADES.update({v: k for k, v in GRADES.items()})


def get_color(color: str) -> str:
    return APPLE_COLORS[color]


def generate_sku(params: dict[str, Any]):
    kind = params['kind']
    color_code = params.pop('color_code', None)
    if params.get('color') and not color_code:
        color_code = APPLE_COLORS[params['color']]
    if not params.get('storage') and params.get('memory'):
        params['storage'] = params['memory']
    match kind:
        case 'iPhone':
            sku = generate_iphone_sku({**params, 'color_code': color_code})
        case 'iPad':
            if params['base'] != 'IPA':
                sku =  '{base}{generation}{screen_inch}{color_code}{storage}'.format(color_code=color_code, **params)
            else:
                sku =  '{base}{generation}{color_code}{storage}'.format(color_code=color_code, **params)
            sku += 'C' if params.get('cellular') else 'W'
            if params.get('restricted'):
                sku += params['restricted']
            if params.get('grade'):
                sku += params['grade']
        case _:
            raise NotImplementedError
    return sku

def generate_iphone_sku(params: dict[str, str]):
    params.update(APPLE_MODELS[params['model']]) # type: ignore
    params.setdefault('generation', '')
    if not params.get('device'):
        params['device'] = params['series']
    if params.get('color') and params.get('storage'):
        sku = '{base}{color_code}{storage}'.format(**params)
        if params.get('restricted'):
            sku += params['restricted']
        if params.get('grade'):
            grade = params['grade']
            params['grade_name'] = GRADES[grade]
            if not grade.isdigit():
                grade = GRADES[grade]
            sku += grade
        return sku
    elif params.get('storage'):
        return '{base}{memory}'.format(**params)
    else:
        return '{base}'.format(**params)

def parse_iphones(
    client: IClient[Any, Any]
):
    return extract.parse_iphones(
        client,
        colors=APPLE_COLORS,
        generate_sku=generate_iphone_sku
    )


def parse_ipads(
    client: IClient[Any, Any]
):
    return extract.parse_ipads(
        client,
        colors=APPLE_COLORS,
        generate_sku=lambda params: '{base}{generation}{screen_inch}{color_code}{memory}'.format(**params)
    )


# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
import sys
from typing import Any

from headless.types import IClient


DEFAULT_HEADERS: dict[str, str] = {
    'User-Agent': "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"
}


report = lambda *x: print(*x, file=sys.stderr)


DESCRIPTION1: set[str] = {
    '661-19620',
    '661-25890',
    '661-25683',
    '923-07174',
    '923-07176',
}


def create_part(model: str, params: dict[str, Any]) -> dict[str, Any]:
    c_attrs: dict[str, Any] = {x['id_countries']: x for x in params['country_specific_attributes']}
    language_attrs: dict[str, Any] = {x['id_languages']: x for x in params['language_specific_attributes']}
    c = c_attrs['BE']
    l = language_attrs['en']
    if model == 'iPhone SE (3rd generation)':
        model = 'iPhone SE 2022'
    name = l['description2'] or l['description1']
    if params['partnumber'] in DESCRIPTION1:
        name = l['description1']
    if str.lower(name) in {str.lower("Display & Screw Kit"), str.lower("Battery & Screw Kit"), str.lower("Camera"), str.lower("Bottom Speaker"), str.lower("Taptic Engine")}:
        name = f'{name} ({model})'
    return {
        'kind': 'Part',
        'name': str.strip(name),
        'namespaces': ['parts.molano.nl'],
        'product_type': 'RAW_MATERIAL',
        'suppliers': ['selfservicerepair.eu'],
        'cost_estimation': str(c['price']),
        'annotations': {
            'selfservicerepair.eu/partnumber': params['partnumber'],
            'selfservicerepair.eu/credit-amount': str(c['creditamount'])
        },
        'search_keys': [
            f"selfservicerepair.eu/partnumber:{params['partnumber']}"
        ]
    }


async def parse_category(
    client: IClient[Any, Any],
    parts: dict[str, Any],
    device_type_id: int,
    model_id: int,
    model_name: str,
    repair_type_id: int,
    repair_type_code: str,
    url: str = "https://selfservicerepair.eu/api/Spares/Search"
):
    response = await client.get(
        url=url,
        headers=DEFAULT_HEADERS,
        params={
            'deviceTypeId': device_type_id,
            'deviceModelId': model_id,
            'repairTypeId': repair_type_id,
            'tools': False,
            'repairTypeCode': repair_type_code
        }
    )
    for dto in await response.json():
        # Ignore bundles.
        part_number: str = dto['partnumber']
        if part_number.startswith('BD') or part_number in parts or not part_number[0].isdigit():
            continue
        parts[part_number] = create_part(model_name, dto)


async def parse_model(
    client: IClient[Any, Any],
    parts: dict[str, Any],
    device_type_id: int,
    model_id: int,
    model_name: str,
    url: str = "https://selfservicerepair.eu/api/RepairTypes/GetByModel"
):
    response = await client.get(
        url=url,
        headers=DEFAULT_HEADERS,
        params={
            'deviceTypeId': device_type_id,
            'modelId': model_id
        }
    )
    for obj in await response.json():
        report(f"- Scraping category {obj['repairtype']}")
        await parse_category(client, parts, device_type_id, model_id, model_name, obj['id'], obj['repairtypecode'])


async def parse_selfservicerepair(client: IClient[Any, Any]) -> list[Any]:
    url: str = "https://selfservicerepair.eu/api/DeviceModels/GetByDeviceType?deviceTypeId=1&modelId=1"
    parts: dict[str, Any] = {}
    response = await client.get(url=url, headers=DEFAULT_HEADERS)
    for obj in await response.json():
        report(f"Scraping {obj['model']}")
        await parse_model(client, parts, obj['id_devicetypes'], obj['id'], obj['model'])
    return list([x[1] for x in sorted(parts.items(), key=lambda x: x[0])])


# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from .const import APPLE_MODELS
from .v2 import APPLE_COLORS


GRADES: dict[str, str] = {
    'A+'    : '0',
    'A'     : '1',
    'B+'    : '2',
    'B'     : '3',
    'C+'    : '4',
    'C'     : '5',
    'D'     : '6',
}


def generate_sku(params: dict[str, Any]) -> str:
    kind = params.get('kind')
    color_code = APPLE_COLORS[params['color']] if params.get('color') else None
    is_bulk = params.get('bulk')
    is_trading = params.get('trading')
    model_suffix: str = ''
    if is_bulk:
        model_suffix = 'B'
    if is_trading:
        model_suffix = 'T'
    match kind:
        case "iPad":
            raise NotImplementedError
        case "iPhone":
            params.update(APPLE_MODELS[params['model']]) # type: ignore
            params.setdefault('generation', '')
            if color_code:
                sku = '{base}{color_code}{storage}'.format(color_code=color_code, **params)
            else:
                sku = '{base}{model_suffix}{storage}'.format(color_code=color_code, model_suffix=model_suffix, **params)
        case _:
            raise NotImplementedError
    if params.get('restricted'):
        sku += params['restricted']
    if params.get('grade'):
        grade: str = params['grade']
        if not grade.isdigit():
            grade = GRADES[grade]
        sku += grade
    return sku


# Copyright (C) 2023 Cochise Ruhulessin
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# type: ignore
import collections
import copy
import itertools
import re
from typing import Any
from typing import Callable
from typing import AsyncIterator

import lxml.html
from headless.types import IClient

from .const import APPLE_MODELS
from .const import APPLE_IDENTIFIERS
from .utils import table_to_list


WIKI_URL: str = "https://www.theiphonewiki.com/wiki/Models"


TAGS: dict[int, str] = {
    1: 'apple:model:exclusive:cn',
    2: 'apple:model:limited:iphone6',
    3: 'apple:model:exclusive:jp',
    4: 'apple:model:exclusive:jp',
    5: 'apple:model:exclusive:cn',
    6: 'apple:model:demo',
    7: 'apple:model:exclusive:cn',
}


MACBOOKS: dict[str, Any] = {
    'MBP2023M2': {
        'name': 'MacBook Pro M2',
        'year': 2023,
        'screen_sizes': [13],
        'colors': [
            "Space Grey",
            "Silver"
        ],
        'memory': [
            "8GB",
            "16GB",
            "24GB",
        ],
        'storage': [
            "256GB",
            "512GB",
            "1TB",
            "2TB"
        ]
    },
    'MPB2023M2P': {
        'year': 2023,
        'cpu': [
            'm2p10x16',
            'm2p12x19',
        ],
        'colors': [
            "Space Grey",
            "Silver"
        ],
        'memory': [
            "16GB",
            "32GB",
        ],
        'storage': [
            "512GB",
            "1TB",
            "2TB",
            "4TB",
            "8TB",
        ]
    }
}


def parse_anumbers(row: Any, values: str) -> list[dict[str, Any]]:
    numbers: list[dict[str, Any]] = [f'A{x}' for x in str.split(values, 'A') if x and x != 'A?']
    for i, number in enumerate(numbers):
        if number == 'A?':
            numbers[i] = None
            continue
        params: dict[str, Any] = {}
        if len(number) != 5:
            m = re.match('(?P<number>.*)\[(?P<footnote>[0-9])\]$', number)
            if m is None:
                raise Exception(number)
            params['tag'] = TAGS[int(m.group('footnote'))]
            number = m.group('number')
        params['number'] = number
        numbers[i] = params
    return list(filter(bool, numbers))

async def parse_iphonewiki(
    client: IClient[Any, Any],
    kind: str,
    url: str = WIKI_URL
) -> AsyncIterator[Any]:
    response = await client.get(url=url)
    root = lxml.html.fromstring(response.content)
    for element in root.findall('.//table'):
        # Get the first row to determine which kind of models the
        # table contains.
        tbody = element[0]
        assert tbody.tag == 'tbody', tbody.tag
        try:
            first_row = tbody[1]
        except IndexError:
            continue
        name = str.strip(first_row[0].text or '')
        if len(first_row[0]) > 0: # Is a hyperlink
            name = str.strip(first_row[0][0].text)
        assert isinstance(name, str)
        if not str.lower(name).startswith(kind):
            continue
        rows = table_to_list(tbody)
        for row in rows[1:]:
            row[8] = [str.strip(x) for x in row[8].split(',')]
            row[1] = parse_anumbers(row, row[1])

            if row[8] == ['?']:
                row[8] = []
            yield row


async def parse_ipads(
    client: IClient[Any, Any],
    colors: dict[str, str],
    generate_sku: Callable[[dict[str, Any]], str] = lambda x: ''
):
    models: list[dict[str, Any]] = []
    model_identifiers: dict[Any, Any] = collections.defaultdict(set)
    async for row in parse_iphonewiki(client, 'ipad'):
        row[7] = str.replace(row[7], ' ', '')
        color_code = colors[row[6]]

        attrs = APPLE_IDENTIFIERS[row[5]]
        attrs.update(copy.deepcopy(APPLE_MODELS[row[0]]))
        msin = f"urn:msin:apple.com:"
        msin += str.lower(attrs.get('spec') or 'xx') + ':dev:ipad:'
        msin += attrs['msin']
        msin += ':' + str.lower(row[7])
        msin += ':' + str.lower(color_code)
        msin += ':' + ('w' if not attrs.get('cellular') else 'c')
        params: dict[str, int | str] = {
            'kind': 'iPad',
            'models': row[1],
            'identifier': row[5],
            'generation': 1,
            'screen_size': '',
            'screen_inch': '',
            **attrs,
            'memory'    : row[7],
            'model'     : row[7],
            'msin'      : msin,
            'series'    : row[0],
            'color'     : row[6],
            'color_code': color_code
        }
        if params.get('screen_size'):
            params['screen_inch'] = params['screen_size'].split('.')[0]
        params['sku'] = generate_sku(params)
        models.append(params)

        # Collect models
        for number in params['models']:
            k = (number['number'], row[5], row[6], row[7])
            model_identifiers[k].update(row[8])

    # Group by A numbers
    versions: dict[tuple[str, str | None], set[str]] = collections.defaultdict(set)
    for params in models:
        identifier = params['identifier']
        for model in params['models']:
            tag = model.get('tag')
            versions[(identifier, tag)].add(model['number'])

    for tag, params in itertools.product([None, *TAGS.values()], models):
        params = copy.deepcopy(params)
        identifier = params['identifier']
        numbers = versions.get((identifier, tag))
        if not numbers:
            continue
        params['tag'] = tag
        params['variants'] = []

        # Collect model numbers
        for a in list(sorted(numbers)):
            k = a, params['identifier'], params['color'], params['memory']
            params['variants'].append({
                'number': a,
                'models': list(sorted(model_identifiers[k]))
            })
        del params['models']
        yield params


async def parse_iphones(
    client: IClient[Any, Any],
    colors: dict[str, str],
    generate_sku: Callable[[dict[str, Any]], str] = lambda x: '',
):
    models: list[dict[str, Any]] = []
    model_identifiers: dict[Any, Any] = collections.defaultdict(set)
    async for row in parse_iphonewiki(client, 'iphone'):
        row[7] = str.replace(row[7], ' ', '')
        params: dict[str, int | str] = {
            'kind'      : 'iPhone',
            'generation': '',
            'models': row[1],
            'identifier': row[5],
            **copy.deepcopy(APPLE_MODELS[row[0]]),
            'memory'    : row[7],
            'model'     : row[0],
            'series'    : row[0],
            'color'     : row[6],
            'color_code': colors[row[6]]
        }
        params['sku'] = generate_sku(params)
        models.append(params)

        # Collect models
        for number in params['models']:
            k = (number['number'], row[5], row[6], row[7], False)
            model_identifiers[k].update(row[8])

    # Group by A numbers
    versions: dict[tuple[str, str | None], set[str]] = collections.defaultdict(set)
    for params in models:
        identifier = params['identifier']
        for model in params['models']:
            tag = model.get('tag')
            versions[(identifier, tag)].add(model['number'])

    seen = set()
    for tag, params in itertools.product([None, *TAGS.values()], models):
        params = copy.deepcopy(params)
        identifier = params['identifier']
        numbers = versions.get((identifier, tag))
        if not numbers:
            continue
        params['tag'] = tag
        params['variants'] = []
        k = (
            params['series'],
            params.get('tag'),
            params['memory'],
            params['color'],
            params['identifier']
        )
        del params['models']
        if k in seen:
            continue
        seen.add(k)

        # Collect model numbers
        for a in list(sorted(numbers)):
            k = a, params['identifier'], params['color'], params['memory'], False
            s = model_identifiers[k]
            for n in s:
                if not n.endswith('[6]'):
                    continue
                s.add(n[:-3])
                s.remove(n)
            params['variants'].append({
                'number': a,
                'model': list(sorted(s))
            })

        if params['memory'] == '32GB[2]' and params['series'] == 'iPhone 6':
            params['memory'] = '32GB'
            params['year'] = 2017
        yield params

#        1: 'apple:model:exclusive:cn',
#    2: 'apple:model:limited:iphone6',
#    3: 'apple:model:exclusive:jp',
#    4: 'apple:model:exclusive:jp',
#    5: 'apple:model:exclusive:cn',
#    6: 'apple:model:demo',
#    7: 'apple:model:exclusive:cn',